// Remember, Mouse is the struct that knows about, for example, the robot's position in the maze.
// This file is only for defining a hardware abstraction layer
// The code in this file will be dirty, it will deal with setting up encoders and all sorts of BS

const map = @import("map.zig");
const Point = map.Point;
const Cardinal = map.Cardinal;

// Hardware-specific declarations
const microzig = @import("microzig");
const rp2040 = microzig.hal;
const time = rp2040.time;
const i2c = rp2040.i2c;
const gpio = rp2040.gpio;

const pin_config = rp2040.pins.GlobalConfiguration {
    // 6 colored LEDs
    .GPIO0 = .{
        .name = "led_4",
        .direction = .out,
    },
    .GPIO1 = .{
        .name = "led_3",
        .direction = .out,
    },
    .GPIO2 = .{
        .name = "led_5",
        .direction = .out,
    },
    .GPIO3 = .{
        .name = "led_2",
        .direction = .out,
    },
    .GPIO4 = .{
        .name = "led_6",
        .direction = .out,
    },
    .GPIO5 = .{
        .name = "led_1",
        .direction = .out,
    },

    // 3 of the 5 LiDAR clock enable pins
    // Instead of writing 1 to the CE pins, we should set them as inputs and change them to pull-high
    // Then we set as an output and write 0
    // This is because the sensor board is actually running at 1.8V
    // Each LiDAR has:
    //      GPIO0 - enable pin, must be pulled high during boot and when using the chip, this is our ce pin
    //      SCL
    //      SDA
    // We don't need to worry about these pins in software, they're wired in!
    //      GPIO1 -
    //      AVDD - needs power second
    //      AVSS_VCSEL
    //      AVDD_VCSEL - needs power first
    // The config here doesn't really matter because we dynamically update these every time we write to them.
    .GPIO6 = .{
        // LCE3
        .name = "lidar_front_left_ce",
        .direction = .out,
    },
    .GPIO7 = .{
        // LCE5
        .name = "lidar_front_ce",
        .direction = .out,
    },
    .GPIO8 = .{
        // LCE1
        .name = "lidar_back_left_ce",
        .direction = .out,
    },

    // Motors
    .GPIO9 = .{
        // M/Fault ("indicates an error condition in the motor controller (such as over current)")
        .name = "motor_fault",
        .direction = .in,
    },

    // Motor Right (or B)
    .GPIO10 = .{
        // MB2
        .name = "motor_right_2",
        .direction = .out,
    },
    .GPIO11 = .{
        // MB1
        .name = "motor_right_1",
        .direction = .out,
    },
    // Motor Left (or A)
    .GPIO12 = .{
        // MA2
        .name = "motor_left_2",
        .direction = .out,
    },
    .GPIO13 = .{
        // MA1
        .name = "motor_left_1",
        .direction = .out,
    },

    // Motor Encoders
    // Zach recommends triggering a double-edged interrupter on 1 and clocking on 2
    // You read the second to know if you're going forward or backwards
    .GPIO14 = .{
        // MAE2
        .name = "motor_left_encoder_1",
        .direction = .in,
    },
    .GPIO15 = .{
        // MAE1
        .name = "motor_left_encoder_2",
        .direction = .in,
    },
    .GPIO16 = .{
        // MBE2
        .name = "motor_right_encoder_2",
        .direction = .in,
    },
    .GPIO17 = .{
        // MBE1
        .name = "motor_right_encoder_1",
        .direction = .in,
    },

    // Buttons
    .GPIO18 = .{
        // BTN1
        .name = "button_1",
        .direction = .in,
    },
    .GPIO19 = .{
        // BTN2
        .name = "button_2",
        .direction = .in,
    },

    // SD Card
    // .GPIO20 = .{
    //     // SD_MISO
    //     .name = "button_2",
    //     .direction = .in,
    // },
    // .GPIO21 = .{
    //     // SD_CSB
    //     .name = "button_2",
    //     .direction = .in,
    // },
    // .GPIO22 = .{
    //     // SD_CLK
    //     .name = "button_2",
    //     .direction = .in,
    // },
    // .GPIO23 = .{
    //     // SD_MOSI
    //     .name = "button_2",
    //     .direction = .in,
    // },

    // LiDAR signals
    // These aren't GPIO pins, they're I2C pins, so we don't need to configure them here
    // .GPIO24 = .{
    //     // SCL
    //     .name = "",
    //     .direction = .out,
    // },
    // .GPIO25 = .{ // This is the on board LED on the Pico
    //     // SDA
    //     .name = "",
    //     .direction = .out,
    // },

    // 2 of the 5 encoders
    .GPIO26 = .{
        // LCE2
        .name = "lidar_back_right_ce",
        .direction = .out,
        // .pull =
    },
    .GPIO27 = .{
        // LCE4
        .name = "lidar_front_right_ce",
        .direction = .out,
    },

    // Motor Current monitoring
    // These pins support Analog-Digital Conversion, which is cool
    // TODO: What do these do? What do we do with them?
    .GPIO28 = .{
        // MAI
        .name = "ma_i",
        .direction = .in,
    },
    .GPIO29 = .{
        // MBI
        .name = "mb_i",
        .direction = .in,
    },
};
// comptime {
//     @compileLog(@typeInfo(@TypeOf(pin_config.apply)));
// }
// var pins: @typeInfo(@TypeOf(pin_config.apply)).Fn.return_type = undefined;

const LEFT_MOTOR = 1;
const RIGHT_MOTOR = 2;

// 254mm square size
const SQUARE_SIZE = 254;

var pins: rp2040.pins.Pins(pin_config) = undefined;

pub fn toggle_led_6 () void {
    // TODO: this does bad things if pins hasn't been defined by the first core yet
    pins.led_5.toggle();
}

// Zach confirms i2c 0
// const i2c0 = i2c.num(0);

// Pin stuff
pub fn setup () void {
    pins = pin_config.apply();

    // const lidar_ce_pins = [_]microzig.hal.gpio.Pin{
    //     pins.lidar_front_left_ce,
    //     pins.lidar_front_right_ce,
    //     pins.lidar_back_right_ce,
    //     pins.lidar_back_left_ce,
    //     pins.lidar_front_ce,
    // };

    // _ = i2c0.apply(.{
    //     .clock_config = rp2040.clock_config, // Zach says the default here is fine
    //     .scl_pin = gpio.num(24),
    //     .sda_pin = gpio.num(25),
    // });

    // // We need to setup the LiDAR addresses
    // // TODO: Create a Lidar struct with methods
    // // The way we're doing this right now, we are enabling every LiDAR
    // const LIDAR_ADDR_BASE: i2c.Address = @enumFromInt(0x50);
    // // Disable all of them
    // for (lidar_ce_pins) |pin| {
    //     pin.set_direction(.out);
    //     pin.put(0);
    // }
    // // One by one, enable and give an address
    // var current_address = LIDAR_ADDR_BASE;
    // for (lidar_ce_pins) |pin| {
    //     // Enable
    //     pin.set_direction(.in);
    //     pin.set_pull(.up);
    //     // Write an address
    //     i2c0.set_address(current_address);
    //     current_address = @enumFromInt(@intFromEnum(current_address) + 1);
    // }


    // pins.onboard_led.toggle();
    // pins.onboard_led.toggle();
    // TODO

    // start_alt_core()
}

pub fn moveForward() void {
    // TODO
}

pub fn turn90(direction: map.TurningDirection) void {
    // TODO
    _ = direction;
}

pub fn turn180() void {
    // TODO
}

// pub fn readFront () !bool
// pub fn readLeft () !bool
// pub fn readRight () !bool
// pub fn readBack () ?? throw I guess
pub fn readRight() bool {
    return true;
}
pub fn readLeft() bool {
    return true;
}
pub fn readFront() bool {
    return true;
}
pub fn readBack() bool {
    return true;
}

// Multicore isn't functional upstream
// const multicore = rp2040.multicore;
// fn alt_core_main_function noreturn {}
pub fn start_alt_core() !void {
    // multicore.launch_core1(alt_core_main_function);
}

pub fn stall() noreturn {

    pins.motor_right_1.put(1);
    pins.motor_right_2.put(0);
    pins.motor_left_1.put(1);
    pins.motor_left_2.put(0);

    time.sleep_ms(5000);

    pins.motor_right_1.put(0);
    pins.motor_right_2.put(0);
    pins.motor_left_1.put(0);
    pins.motor_left_2.put(0);

    while (true) {
        // pins.onboard_led.toggle();

        // const b1 = pins.button_1.read();
        // const b2 = pins.button_2.read();

        // pins.led_1.put(b1);
        // pins.led_6.put(b2);



        time.sleep_ms(10);



        // pins.led_1.put(0);
        // pins.led_2.put(1);
        // time.sleep_ms(500);
        // pins.led_2.put(0);
        // pins.led_3.put(1);
        // time.sleep_ms(500);
        // pins.led_3.put(0);
        // pins.led_4.put(1);
        // time.sleep_ms(500);
        // pins.led_4.put(0);
        // pins.led_6.put(1);
        // time.sleep_ms(500);
        // pins.led_6.put(0);
    }
}

// --- Functions not implemented for the IRL bot  ---
pub fn writeCellScore(p: Point, score: usize) void {
    _ = p;
    _ = score;
    // stdout.print("setText {d} {d} {d}\n", .{p.x, p.y, score}) catch unreachable;
}
pub fn recordWall(p: Point, direction: Cardinal) void {
    _ = p;
    _ = direction;
}
